require 'tk'

#
# 10/16/13 [HET]: Altered to add error message 2 to tell user that previous version of document may not
# contain boxes.
#
# 06/08/15 [HET]: Altered read_prev_file(file) to read just boxes from the .XML file rather than looking for
# a particular type of box, thus removing dependancy on any new types of boxes added. An .XML file only lists
# a particular box once so we can read it this way.
#
# 06/16/15 [HET]: Now looking for the previous version's doc.xml file in the user's work folder which is on
# their computer. We are moving away from using the DataCenter so this change is necessary.
#
# 04/27/16 [htruax]: Changed read_cur_file( ) to look for '# ***** Calculations *****' as an indicator to stop
# reading '# ***** Field Data *****' boxes. The '# ***** Field Data *****' section has been moved from just before the
# Profiles to just after them.
#
# http://stackoverflow.com/questions/856891/unzip-zip-tar-tag-gz-files-with-ruby
#~ require 'rubygems/package'
#~ require 'zlib'
#

#
# #########################################################
#
# This provides a list of boxes that are added to and deleted from the document being worked on as compared
# to the previous version of the document. This tool is started from the SciTE Tools menu and compares the
# Field Data box list in the current document's Ruby file in SciTE to the previous version's XML boxes
# stored on the User's computer [in their work folder].
# 
# #########################################################
#

#
# #########################################################
#
# Message Box section:
#

require 'Win32API'

#
# Button styles:
#
$MB_OK = 0
$MB_OKCANCEL = 1
$MB_ABORTRETRYIGNORE = 2
$MB_YESNOCANCEL = 3
$MB_YESNO = 4
$MB_RETRYCANCEL = 5

#
# Icon styles:
#
$MB_ICONHAND = 16
$MB_ICONQUESTION = 32
$MB_ICONEXCLAMATION = 48
$MB_ICONASTERISK = 64

#
# Default button:
#
$MB_DEFBUTTON1 = 0
$MB_DEFBUTTON2 = 256
$MB_DEFBUTTON3 = 512

#
# Return values:
#
$IDOK = 1
$IDCANCEL = 2
$IDABORT = 3
$IDRETRY = 4
$IDIGNORE = 5
$IDYES = 6
$IDNO = 7

def messageBox(message, title = 'Message', style = 0)
	mb = Win32API.new("user32", "MessageBox", ['L', 'P', 'P', 'L'], 'L')
	return mb.Call(0, message, title, style)
end

#
# #########################################################
#
# Window drawing section.
#
# #########################################################
#

#
# ##################################################
#
# For displaying the title area:
#
ver_num = 1.0

#
# ##################################################
#
# Frames and Labels:
#

font_name = "arial 10"

root = TkRoot.new() { title "BOXES ADDED AND DELETED - Version #{ver_num}" }
root.geometry('+150+500')

wndw_width = 90

prev_label = TkLabel.new(root, :relief=>:sunken, :background=>"lightgrey", :text=>"Previous: ", "font"=>"arial 10 bold", :width=>wndw_width).pack(:fill=>:x, :padx=>5)
cur_label = TkLabel.new(root, :relief=>:sunken, :background=>"lightgrey", :text=>"Current: ", "font"=>"arial 10 bold", :width=>wndw_width).pack(:fill=>:x, :padx=>5)

output_frame = TkFrame.new(root, :borderwidth=>3, :relief=>:ridge).pack(:side=>:top, :fill=>:x, :padx=>5, :pady=>5)

output_label = TkLabel.new(output_frame, :text=>"<Ctrl></> Selects all text.", "font"=>font_name).pack()

out_text_bar_vert = TkScrollbar.new(output_frame).pack(:side=>:right, :fill=>:y)
out_text = TkText.new(output_frame, :height=>6, :width=>wndw_width, "font"=>font_name, :state=>"normal").pack(:fill=>:x, :expand=>true)
out_text.yscrollbar(out_text_bar_vert)

#
# #########################################################
#
# Globals:
#
cur_dir = "C:/temp/work"

#
# This is a quick adjustment to var 'user' for Sherry Hughe's computer. Form some reason, her C:/Users/
# full path is not "C:/Users/shughes/AppData/Local/Temp/DataCenter" as expected, but is 
# instead "C:/Users/shughes.CORP/AppData/Local/Temp/DataCenter".
#
# 06/16/15 [htruax]: Removed as we are now looking at the standard work directory off from the temp folder.
#~ user = ENV['USERNAME']
#~ user += ".CORP" if (user.downcase.eql?("shughes"))

cur_file = ""
prev_file = ""

#
# #########################################################
#
# Methods:
#
def process_errors(err_type, err_num, var1, var2)

	err_msg = [
		"  Previous XML file for #{var1} not found:\n\nFirst make sure the previous file had boxes. If it did not, an XML file will not be created. Exit from this application as all boxes on new document are new.\n\nIf previous file has boxes, do the following:\n\n1) Exit from this application.\n2) Export the previous version's XML file from SuperBoxing.\n3) Restart this application from the #{var2} file buffer in SciTE.",
		"  Application was started from a SciTE buffer that did not contain the Ruby code from a document.\n\nDo the following:\n\n1) Exit from this application.\n2) Open the Current document in SuperBoxing.\n3) Click on the Calc Editor icon in SuperBoxing.\n4) Restart this application from the Current file buffer in SciTE."
	]

	#
	# Using bitwise operations, err_num is set on a bit by bit basis, where each bit represents an error message.
	#
	if (err_num > 0)

		mask = 1

		#
		# Math.frexp(err_num) returns a two element array containing the normalized fraction (a Float) and exponent
		# (a Fixnum) of err_num. In our case, the fraction is 0.5 and the exponent is a power of 2, as our err_num is 1 x 2
		# raised to 0,1,2,3,... So this returns 0.5 x 2[a power]. So, say err_num = 4, which is 1 x 2[2], which is the 3rd
		# error in our list. .frexp returns [0.5, 3] - we want the 3 [3rd error], we use 3 - 1 since we are starting
		# from 0 not 1. Just doing this so that I do not have to keep updating the 0.upto( ) each time I add a new err_num.
		#
		0.upto(Math.frexp(err_num)[1]-1) do |shift|
			if (((mask << shift) & err_num) > 0)
				messageBox(err_msg[shift], err_type, $MB_ICONEXCLAMATION|$MB_OK)
			end
		end
	end
end

#
# This reads in the Field Data boxes from the current Ruby Code file.
#
def read_cur_file(file)

	box_array = Array.new
	process = false

	File.open(file, "r") do |src_file|

		while (line = src_file.gets)

			case 
				when (line.delete("\s\t") =~ /\A#\*\*\*\*\*FieldData\*\*\*\*\*\Z/): process = true

				when (process)

					case (line.delete("\s\t"))

						when (/^(box\d+_\d+)=/): box_array.push($1)

						when (/\A#\*\*\*\*\*Calculations\*\*\*\*\*/)

							process = false
							break
					end
			end
		end
	end

	return box_array
end

#
# This looks for the previous version of the document's XML file on the C: drive of the 'user's computer, in their
# work folder. If located, the file name and path is returned, else nil is returned.
#
def find_prev_file(cur_file)

	prev_file = nil
	found = false

	#
	# Just the file name with no path info - i.e. "wxyz1234.rbw"
	#
	cur_file_name = File.basename(cur_file)

	#
	# The directory path from where they started this application, which should be the working directory off from the temp folder
	# on drive C:.
	#
	cur_dir = File.dirname(cur_file)

	#
	# We must find the xml file in the same folder as cur_file_name is in, and it must start with these 4 characters else
	# we have an error.
	#
	doc_id = cur_file_name[0,4].upcase

	Dir.chdir(cur_dir) do

		#
		# There should be only 1 .xml file in this directory, but being cautious here, plus this routine returns an array.
		# Tests have shown that case is not important as far as Dir.glob() is concerned.
		#
		Dir.glob("#{doc_id}[0-9][0-9][0-9][0-9].xml").each do |xml_file|

			if (doc_id.eql?(xml_file[0,4].upcase))

				prev_file = File.expand_path(xml_file)
				found = true
				break
			end
		end
	end

	process_errors("Error!", 1, "#{cur_file_name[0,8]}", nil) unless (found)
	prev_file
end

#
# This reads in ALL the boxes from the previous XML file, even the barcode boxes. This version of the application requires the
# user to first export the XML file from the previous version. The previous version is started from the WorkFlow system,
# and the previous version is now automatically started from SuperBoxing, from the user's working folder.
#
def read_prev_file(file)

	box_array = Array.new
	process = false

	File.open(file, "r") do |src_file|

		while (line = src_file.gets)
			box_array.push($1) if (line =~ /"(box\d+_\d+)"/)
		end
	end

	box_array
end

#
# Returns a string which indicates which boxes are added or deleted based on fcn: "ADD" or "DEL".
# Boxes that are added/deleted with contiguous numbering [box1_2, box1_3, box1_4] are indicated
# as box1_2-4, single boxes are spelled out as in box1_10 for example.
#
def make_add_del_str(tmp_array, fcn)

	rpt_array = Array.new
	rpt_text = String.new

	start_index = stop_index = page = num = 0

	rpt_text = (fcn.eql?("ADD") ? "Added: " : "Deleted: ")

	if (tmp_array.empty?)
		rpt_text += "None"
	else
		tmp_array.each_index do |i|
	
			if (i == 0)
				page = tmp_array[i][0]
				num = tmp_array[i][1]
			else
				if (tmp_array[i][0] == page)
		
					if (tmp_array[i][1] == num + 1)
		
						num += 1
						stop_index = i
					else
						#
						# End of contiguous box numbering-i.e.box1_2, box1_3, box1_4, store tmp_array start and stop indices.
						#
						rpt_array.push([start_index,stop_index])
						start_index = stop_index = i
						num = tmp_array[i][1]
					end
				else
					#
					# Page is different.
					#
					rpt_array.push([start_index,stop_index])
					start_index = stop_index = i
					page = tmp_array[i][0]
					num = tmp_array[i][1]
				end
			end
		end
	
		rpt_array.push([start_index,stop_index]) unless (tmp_array.empty?)
	
		rpt_array.each do |range|
		
			rpt_text += if (range[0] == range[1])
				"box#{tmp_array[range[0]][0]}_#{tmp_array[range[0]][1]}, "
			else
				"box#{tmp_array[range[0]][0]}_#{tmp_array[range[0]][1]}-#{tmp_array[range[1]][1]}, "
			end
		end
	end

	rpt_text.strip.chomp(",")
end

#
# This produces the boxes added/deleted report for insertion into WorkFlow notes.
#
def make_report(cur_box_array, prev_box_array, out_text)

	tmp_array = add_box_array = del_box_array = Array.new

	add_box_array = cur_box_array - prev_box_array
	del_box_array = prev_box_array - cur_box_array

	add_box_array.each do |box| box.scan(/box(\d+)_(\d+)/) { tmp_array.push([$1.to_i,$2.to_i]) } end
	tmp_array.sort!

	rpt_str = make_add_del_str(tmp_array, "ADD")
	out_text.insert("end", rpt_str)

	tmp_array.clear
	del_box_array.each do |box| box.scan(/box(\d+)_(\d+)/) { tmp_array.push([$1.to_i,$2.to_i]) } end
	tmp_array.sort!

	rpt_str = make_add_del_str(tmp_array, "DEL")
	out_text.insert("end", "\n#{rpt_str}")
end

#~ def get_file_name(dir, type)

	#~ ftypes = case (type)
			#~ when ("RBW"): [ ["Ruby source files", '.rbw'] ]
			#~ when ("XML"): [ ["XML files", '.xml'] ]
		#~ end

	#~ Tk.getOpenFile(:filetypes=>ftypes, :initialdir=>dir)
#~ end

#
# ###########################################################
#
# Menu:
#
#		Read: Currently not used.
#			Read Ruby Code file.
#
#		Exit:
#			Exit the Application.
#
main_menu = TkMenu.new(root)

#~ read_menu = TkMenu.new(main_menu, :tearoff=>false)

#~ #
#~ #
#~ # This reads the .RBW file and builds an array of Box objects, setting each object to the
#~ # box number and the width is not set by this routine [XML file is needed for this].
#~ #
#~ # The form_box_list is then poulated with the box numbers.
#~ #
#~ read_menu.add('command', :label=> "Ruby Source Code File", :command=> proc {

	#~ fname = get_file_name(cur_dir, "RBW")

	#~ if (!fname.empty?)

		#~ cur_dir = File.dirname(fname)
		#~ form_name = fname
		#~ file_label.configure("text", fname)
	#~ end
#~ })

#~ main_menu.add('cascade', :menu=>read_menu, :label=>"Read")

main_menu.add('command', :label=> "Exit", :command=> proc{exit})

#
# Adding main_menu to the root window.
#
root.menu(main_menu)

#
# This is the default file when called from the SciTE Tools menu. It is expected that this application is started from
# the current version in SciTE and the previous version has been started from the DataCenter via SuperBoxing.
#
# When a document is started in SuperBoxing from the DataCenter, the documents xml file [doc.xml] is copied [along
# with various other files] to a temp directory on the user's C: drive. If there was another doc.xml stored on the user's
# C: drive from a previous startup, it will first be deleted so that there is only one doc.xml on the C: drive temporary
# folder at a time.
#
if (ARGV[0])

	#
	# We need to ensure this application was started from a document's Ruby code file i.e. has a file name
	# in AAAAxxxx.rbw where A is any alpha character and x is any digit.
	#
	if (File.basename(ARGV[0]) =~ /([a-zA-Z]{4})\d{4}\.rbw/)

		cur_label.configure("text", "Current:  #{ARGV[0]}")

		#
		# This is the whole thing - path and file name.
		#
		cur_file = ARGV[0]

		#
		# Just the file name with no path info - i.e. "wxyz1234.rbw"
		#
		cur_file_name = File.basename(ARGV[0])

		#
		# The directory path from where they started this application.
		#
		cur_dir = File.dirname(ARGV[0])

		#
		# Now process the files.
		#
		cur_box_array = read_cur_file(cur_file)
		prev_file = find_prev_file(cur_file)

		if (prev_file)

			prev_label.configure("text", "Previous:  #{prev_file}")
			prev_box_array = read_prev_file(prev_file)
			make_report(cur_box_array, prev_box_array, out_text)
		else
			prev_label.configure("text", "Previous: FILE NOT FOUND")
		end
	else
		process_errors("Error!", 2, nil, nil)
	end
end

Tk.mainloop()